/*======================================================================================================================
  (c) Copyright OMRON Europe BV 2016
  All Right Reserved
  OMRON Development & Engineering
  -----------------------------------------------------------------------------------------------------------------*//**
  @file
  
  This is an implementation of the AES128 algorithm, specifically ECB and CBC mode.

  The implementation is verified against the test vectors in:
  National Institute of Standards and Technology Special Publication 800-38A 2001 ED

  ECB-AES128
  ----------

  plain-text:
  6bc1bee22e409f96e93d7e117393172a
  ae2d8a571e03ac9c9eb76fac45af8e51
  30c81c46a35ce411e5fbc1191a0a52ef
  f69f2445df4f9b17ad2b417be66c3710

  key:
  2b7e151628aed2a6abf7158809cf4f3c

  resulting cipher
  3ad77bb40d7a3660a89ecaf32466ef97
  f5d3d58503b9699de785895a96fdbaaf
  43b1cd7f598ece23881b00e3ed030688
  7b0c785e27e8ad3f8223207104725dd4


  NOTE:   String length must be evenly divisible by 16byte (str_len % 16 == 0)
  You should pad the end of the string with zeros if this is not the case.

  ORIGINAL SOURCE: https://github.com/kokke/tiny-AES128-C

  Only changed to conform to Omron Coding Style and OSB (Omron Secure Boot) data types.
  
  @pre -
*//*==================================================================================================================*/

#ifndef EXTENSIONS_AES_INC_AES_H_
#define EXTENSIONS_AES_INC_AES_H_

/*=== Include files ==================================================================================================*/
#include "types.h"

/*=== Constant definitions ===========================================================================================*/
/**
 * Key length in bytes [128 bit]
 */
#define AES128_KEYLEN      (16)

#define CBC             (1)
#define ECB             (1)

/*=== Macro definitions ==============================================================================================*/

/*=== Type definitions for events ====================================================================================*/

/*=== Type definitions ===============================================================================================*/

/*=== Macro-function definitions =====================================================================================*/

/*=== Inline-function definitions ====================================================================================*/

/*=== Function declarations (Interface functions) ====================================================================*/
#if defined(ECB) && ECB
void aes_ecb_encrypt(void const * const input, const t_U8* const enckey, void* const output);
void aes_ecb_decrypt(void const * const input, const t_U8* const deckey, void* const output);
#endif //#if defined(ECB) && ECB

#if defined(CBC) && CBC
void aes_cbc_encrypt(void* output, void* input, t_U32 const length, const t_U8* enckey, t_U8* const iv);
void aes_cbc_decrypt(void* output, void* input, t_U32 const length, const t_U8* deckey, t_U8* const iv);
#endif //#if defined(CBC) && CBC

#endif
