/* =====================================================================================================================
 (c) Copyright OMRON Europe BV 2016
 All Right Reserved
 OMRON Development & Engineering
 ================================================================================================================== *//**
 @file

  This file is part of Omron Secure Boot (OSB).

  Omron Secure Boot (OSB) is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Omron Secure Boot (OSB) is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Omron Secure Boot (OSB).  If not, see <http://www.gnu.org/licenses/>.

 Implements the application logic for storing the RSA key in the TPM and to retrieve it when verifying data

 @internal
 *//* ================================================================================================================ */

/* ================ Global include files ============================================================================ */
#include "app_rsa_custom.h"
#include "bigdigits.h"

/* ================ Sub-system include files ======================================================================== */

/* ================ Local constant definitions ====================================================================== */

/* ================ Local macro definitions ========================================================================= */

/* ================ Local type definitions ========================================================================== */

/* ================ Local data definitions ========================================================================== */

/* ================ Local functions ================================================================================= */

/* ================ Global functions (Sub-system scope) ============================================================= */

/* ================ Global functions (Global system scope) ======================================================= */
/**
 * Verify signed data
 * @param input buffer of data to be verified
 * @param inputLen [in] length of input buffer in bytes
 * @param output verified data (encrypted with public key)
 * @param mod [in] The RSA modulus. Should have the same length as inputLen
 * @param pub [in] The RSA public key. Should have the same length as inputLen
 * @param keyLen [in] The key-length in bytes. Should be the same as inputLen
 * @return EXIT code indicating if verification was successfully
 */
int apprsacstm_encrypt(t_U8* const input, int const inputLen, t_U8* const output, t_U8 const * const mod, t_U8 const * const pub, int const keyLen)
{
    int status = OMN_EXIT_FAILURE;

    if (((inputLen % BYTES_PER_DIGIT)  == 0) && (inputLen == keyLen)) {
        status = mpModExp((DIGIT_T*)output, (DIGIT_T*)input, (DIGIT_T*)pub, (DIGIT_T*)mod, inputLen / BYTES_PER_DIGIT);
    }
    return status;
}

/* ================ End of file ===================================================================================== */
