/* =====================================================================================================================
(c) Copyright OMRON Europe BV 2016
All Right Reserved 
OMRON Development & Engineering 
================================================================================================================== *//**
@file 

  This file is part of Omron Secure Boot (OSB).

  Omron Secure Boot (OSB) is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Omron Secure Boot (OSB) is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Omron Secure Boot (OSB).  If not, see <http://www.gnu.org/licenses/>.

<short description of the file in one sentence>. 

[Optional extended description of the file. More details] 

@internal
*//* ================================================================================================================ */ 

/* ================ Global include files ============================================================================ */
#include "convars.h"
#include "shared.h"

/* ================ Sub-system include files ======================================================================== */

/* ================ Local constant definitions ====================================================================== */
#define MAX_VAR_CNT         (32u)
#define MAX_VAR_NAME_CNT    (32u)
#define MAX_VAR_VALUE_CNT   (128u)

/* ================ Local macro definitions ========================================================================= */

/* ================ Local type definitions ========================================================================== */
typedef struct {
    char    varIdent[MAX_VAR_NAME_CNT + 1u];
    char    varValue[MAX_VAR_VALUE_CNT + 1u];
    t_bool  isReadOnly;
    t_bool  inUse;
} T_VAR_ENTRY;
/* ================ Local data definitions ========================================================================== */

static struct {
    T_VAR_ENTRY vars[MAX_VAR_CNT];
    t_U16       listCounter;
} convars;

/* ================ Local functions ================================================================================= */
/**
 *
 * @param varIdent
 * @param varLength
 * @return
 */
static char const * getVarName(char const * const varIdent, t_U8* const varLength)
{
    char const * result = NULL;

    if (varLength != NULL) {
        char const * const varIdentStart = varIdent;

        *varLength = 0;
        while ((varIdentStart[*varLength] != ' ' ) &&
               (varIdentStart[*varLength] != 0   ) &&
               (varIdentStart[*varLength] != '\n') &&
               (varIdentStart[*varLength] != '\r') &&
               (varIdentStart[*varLength] != '\t') &&
               (varIdentStart[*varLength] != ';' ) &&
               (*varLength                 < (((t_U16)(1u << (sizeof(*varLength) * 8u))) - 1u))) {
            (*varLength)++;
        }
        result = varIdentStart;
    }
    return result;
}

/**
 *
 * @param varName
 * @param varLength
 * @param convarIdent
 * @return
 */
static t_bool cmpIdent(char const * const varName, t_U8 const varLength, char const * const convarIdent)
{
    unsigned idx = 0;
    t_bool   isSame;

    if (varLength != 0) {
        while ((varName[idx] == convarIdent[idx]) && (idx < varLength)) {
            idx++;
        }

        if ((idx == varLength) && (convarIdent[idx] == '\0')) {
            isSame = TRUE;
        } else {
            isSame = FALSE;
        }
    }

    return isSame;
}

/**
 *
 * @param varIdent
 * @return
 */
static T_VAR_ENTRY* getVarEntry(char const * const varIdent)
{
    T_VAR_ENTRY*       emptyEntry   = NULL;
    T_VAR_ENTRY*       entry        = NULL;
    t_bool             isFound      = FALSE;
    unsigned           idx          = 0;
    t_U8               varLength;
    char const * const varName = getVarName(varIdent, &varLength);

    if (varLength > 0) {
        while (idx < ARRAY_SIZE(convars.vars) && (ISFALSE(isFound))) {
            T_VAR_ENTRY* const convarEntry = &convars.vars[idx];

            if (ISTRUE(convarEntry->inUse)) {
                isFound = cmpIdent(varName, varLength, convarEntry->varIdent);
                if (ISTRUE(isFound)) {
                    entry = convarEntry;
                }
            } else if (emptyEntry == NULL) {
                emptyEntry = convarEntry;
            }
            idx++;
        }
    }

    return (T_VAR_ENTRY*)((entry != NULL) ? entry : emptyEntry);
}

/**
 *
 * @param varEntry
 * @param src
 * @return
 */
static t_bool copyVarIdent(T_VAR_ENTRY* const varEntry, char const * const src)
{
    t_bool             hasError = TRUE;
    t_U8               varLength;
    char const * const varIdent = getVarName(src, &varLength);

    if ((varEntry != NULL) && (varLength > 0)) {
        if (varLength < sizeof(varEntry->varIdent) - 1) {
            (void)memcpy(varEntry->varIdent, varIdent, varLength);
            varEntry->varIdent[varLength] = '\0';
            hasError                      = FALSE;
        }
    }
    return hasError;
}

/**
 *
 * @param varEntry
 * @param src
 */
static t_bool copyVarStrValue(T_VAR_ENTRY* const varEntry, char const * const src)
{
    t_bool hasError = TRUE;

    if ((varEntry != NULL) && (src != NULL)) {
        unsigned idx = 0;

        while ((idx < (sizeof(varEntry->varValue)-1))) {
            varEntry->varValue[idx] = src[idx];
            idx++;
        }
        if (idx <= (sizeof(varEntry->varValue)-1)) {
            varEntry->varValue[idx] = '\0';
            hasError = FALSE;
        }
    }
    return hasError;
}

/**
 *
 * @param dst
 * @param dstLength
 * @param src
 * @param srcLength
 * @return
 */
static t_bool convToStr(char* const dst, t_U16 const dstLength, t_U8 const * const src, t_U16 const srcLength)
{
    t_bool hasError = TRUE;

    if (((srcLength * 2) + 1) <= dstLength) {
        unsigned idx;

        for (idx = 0; idx < srcLength; idx++) {
            t_U8 const upperNible = (t_U8)((src[idx] >> 4) & 0x0Fu);
            t_U8 const lowerNible = (t_U8)(src[idx] & 0x0Fu);

            if (upperNible <= 9) {
                dst[idx*2] = '0' + upperNible;
            } else {
                dst[idx*2] = 'A' + (upperNible - 10);
            }
            if (lowerNible <= 9) {
                dst[idx*2+1] = '0' + lowerNible;
            } else {
                dst[idx*2+1] = 'A' + (lowerNible - 10);
            }
        }
        dst[idx*2] = '\0';
        hasError = FALSE;
    }

    return hasError;
}

/* ================ Global functions (Sub-system scope) ============================================================= */

/* ================ Global functions (Global system scope) ======================================================= *//**
@endinternal
*//* ================================================================================================================ */
void convars_init(void)
{
    (void)memset(&convars, 0, sizeof(convars));
}

t_bool convars_isVar(t_U8 const * const cmdlChar)
{

    t_bool isVar = FALSE;

    if (cmdlChar != NULL) {
        if ((*cmdlChar == VAR_IDENT) &&
            (*(cmdlChar+1) != ' ')   &&
            (*(cmdlChar+1) != '\n')  &&
            (*(cmdlChar+1) != '\r')  &&
            (*(cmdlChar+1) != '\t')  &&
            (*(cmdlChar+1) != ';')   &&
            (*(cmdlChar+1) != '\0')) {
            isVar = TRUE;
        }
    }

    return isVar;
}

/**
 *
 * @param cmdlVarIdent
 * @param cmdlVarStrVal
 * @param asReadOnly
 * @param override
 * @return
 */
T_CONVAR_RESULT convars_defVarStr(char const * const cmdlVarIdent,
                                  char const * const cmdlVarStrVal,
                                  t_bool const       asReadOnly,
                                  t_bool const       override)
{
    T_CONVAR_RESULT result = CONVAR_PARAM_ERROR;

    if ((cmdlVarIdent != NULL) && (cmdlVarStrVal != NULL) && (*cmdlVarIdent == VAR_IDENT)) {
        T_VAR_ENTRY* const convar = getVarEntry(cmdlVarIdent);

        if (convar != NULL) {
            if (ISFALSE(convar->inUse)) {
                t_bool const hasError = copyVarIdent(convar, cmdlVarIdent);

                if (ISFALSE(hasError)) {
                    t_bool const hasCopyError = copyVarStrValue(convar, cmdlVarStrVal);

                    if (ISFALSE(hasCopyError)) {
                        convar->isReadOnly  = asReadOnly;
                        convar->inUse       = TRUE;
                        result              = CONVAR_OK;
                    } else {
                        result = CONVAR_VAL_WONT_FIT;
                    }
                } else {
                    result = CONVAR_PARAM_ERROR;
                }
            } else {
                if (ISFALSE(convar->isReadOnly) || (ISTRUE(override))) {
                    t_bool const hasCopyError = copyVarStrValue(convar, cmdlVarStrVal);

                    if (ISFALSE(hasCopyError)) {
                        convar->isReadOnly  = asReadOnly;
                        result              = CONVAR_OK;
                    } else {
                        result = CONVAR_VAL_WONT_FIT;
                    }
                } else {
                    result = CONVAR_RO;
                }
            }
        } else {
            result = CONVAR_FULL;
        }
    }
    return result;
}

/**
 *
 * @param cmdlVarIdent
 * @param cmdlVarBinVal
 * @param cmdlVarBinValLength
 * @param asReadOnly
 * @param override
 * @return
 */
T_CONVAR_RESULT convars_defVarBin(char const * const  cmdlVarIdent,
                                  t_U8 const * const  cmdlVarBinVal,
                                  t_U16 const         cmdlVarBinValLength,
                                  t_bool const        asReadOnly,
                                  t_bool const        override)
{
    T_CONVAR_RESULT result = CONVAR_PARAM_ERROR;

    if ((cmdlVarIdent        != NULL)   &&
        (cmdlVarBinVal       != NULL)   &&
        (cmdlVarBinValLength  > 0)      &&
        (*cmdlVarIdent        == VAR_IDENT)) {
        T_VAR_ENTRY* const convar = getVarEntry(cmdlVarIdent);

        if (convar != NULL) {
            if (ISFALSE(convar->inUse)) {
                t_bool const hasError = copyVarIdent(convar, cmdlVarIdent);

                if (ISFALSE(hasError)) {
                    t_bool const hasConvError = convToStr(convar->varValue,
                                                          (t_U16)sizeof(convar->varValue),
                                                          cmdlVarBinVal,
                                                          cmdlVarBinValLength);

                    if (ISFALSE(hasConvError)) {
                        convar->isReadOnly  = asReadOnly;
                        convar->inUse       = TRUE;
                        result              = CONVAR_OK;
                    } else {
                        result = CONVAR_VAL_WONT_FIT;
                    }
                } else {
                    result = CONVAR_PARAM_ERROR;
                }
            } else {
                if (ISFALSE(convar->isReadOnly) || (ISTRUE(override))) {
                    t_bool const hasConvError = convToStr(convar->varValue,
                                                          (t_U16)sizeof(convar->varValue),
                                                          cmdlVarBinVal,
                                                          cmdlVarBinValLength);

                    if (ISFALSE(hasConvError)) {
                        convar->isReadOnly  = asReadOnly;
                        result              = CONVAR_OK;
                    } else {
                        result              = CONVAR_VAL_WONT_FIT;
                    }
                } else {
                    result = CONVAR_RO;
                }
            }
        } else {
            result = CONVAR_FULL;
        }
    }
    return result;
}

T_CONVAR_RESULT convars_udefVar(char const * const cmdlVarIdent)
{
    T_CONVAR_RESULT result = CONVAR_PARAM_ERROR;

    if ((cmdlVarIdent != NULL) && (*cmdlVarIdent == VAR_IDENT)) {
        T_VAR_ENTRY* const convar = getVarEntry(cmdlVarIdent);

        if (ISTRUE(convar->inUse)) {
            if (ISFALSE(convar->isReadOnly)) {
                convar->inUse = FALSE;
                result        = CONVAR_OK;
            } else {
                result        = CONVAR_RO;
            }
        } else {
            result            = CONVAR_NOT_FOUND;
        }
    } else {
        result                = CONVAR_INTERNAL_ERROR;
    }
    return result;
}

/**
 *
 * @param cmdlVarIdent
 * @param varStrValue
 * @return
 */
T_CONVAR_RESULT convars_getVar(char const * const cmdlVarIdent, char** const varStrValue)
{
    T_CONVAR_RESULT result = CONVAR_PARAM_ERROR;

    if ((cmdlVarIdent != NULL) && (varStrValue != NULL) && (*cmdlVarIdent == VAR_IDENT)) {
        T_VAR_ENTRY* const convar = getVarEntry(cmdlVarIdent);

        if (ISTRUE(convar->inUse)) {
            *varStrValue = convar->varValue;
            result = CONVAR_OK;
        } else {
            result = CONVAR_NOT_FOUND;
        }
    } else {
        result     = CONVAR_INTERNAL_ERROR;
    }

    return result;
}

/**
 *
 * @param cmdlVarIdent
 * @param varIdentLength
 * @return
 */
char* convars_getVarIdent(char const * const cmdlVarIdent, t_U8* const varIdentLength)
{
    char const* result;

    if (*cmdlVarIdent == VAR_IDENT) {
        result = getVarName(cmdlVarIdent, varIdentLength);
    } else {
        if (varIdentLength != NULL) {
            *varIdentLength = 0;
        }
        result          = NULL;
    }
    return result;
}

/**
 *
 */
void convars_listReset(void)
{
    convars.listCounter = 0;
}

/**
 *
 * @param varIdent
 * @param varValue
 * @param isReadOnly
 * @return
 */
t_bool convars_listNextItem(char** varIdent, char** varValue, t_bool* const isReadOnly)
{
    t_bool hasItem = FALSE;

    while (ISFALSE(convars.vars[convars.listCounter].inUse) && (convars.listCounter < ARRAY_SIZE(convars.vars))) {
        convars.listCounter++;
    }

    if ((convars.listCounter < ARRAY_SIZE(convars.vars)) && ISTRUE(convars.vars[convars.listCounter].inUse)) {
        *varIdent   = convars.vars[convars.listCounter].varIdent;
        *varValue   = convars.vars[convars.listCounter].varValue;
        *isReadOnly = convars.vars[convars.listCounter].isReadOnly;
        convars.listCounter++;
        hasItem     = TRUE;
    }

    return hasItem;
}
/* ================ End of file ===================================================================================== */
