/*======================================================================================================================
 (c) Copyright OMRON Europe BV 2016
 All Right Reserved
 OMRON Development & Engineering
 -----------------------------------------------------------------------------------------------------------------*//**
 @file

 Implements the interface functions to access the TPM
 *//*==================================================================================================================*/

#ifndef EXTENSIONS_TPM_INC_TPM_DRIVER_H_
#define EXTENSIONS_TPM_INC_TPM_DRIVER_H_

/*=== Include files ==================================================================================================*/

#include "utils.h"
#include "types.h"
#include "tpm.h"

/*=== Constant definitions ===========================================================================================*/

// pcr size in bytes
#define PCR_SIZE                (20)

// # of pcr indices
#define TPM_PCRS                (16)

// area of TPM used for key storage
#define TPM_RSA0_2_AREA         ((t_U32)0x1000a000)
#define TPM_RSA3_5_AREA         ((t_U32)0x1000b000)
#define TPM_AES_AREA            ((t_U32)0x1000c000)

#define NR_RSA_KEYS             (6)

// reading requires NVRAM area authorization
#define TPM_PERM_AUTHREAD       ((t_U8)0x00)
// writing requires NVRAM area authorization
#define TPM_PERM_AUTHWRITE      ((t_U8)0x01)
// writing requires owner authorization
#define TPM_PERM_OWNERREAD      ((t_U8)0x02)
// reading requires owner authorization
#define TPM_PERM_OWNERWRITE     ((t_U8)0x03)
// writing requires physical presence
#define TPM_PERM_PPREAD         ((t_U8)0x04)
// reading requires physical presence
#define TPM_PERM_PPWRITE        ((t_U8)0x05)
// write to index 0 locks the NVRAM area until TPM_STARTUP(ST_CLEAR)
#define TPM_PERM_GLOBALLOCK     ((t_U8)0x06)

// tpm vendor identifiers
#define TPM_VEN_AMD             ((t_U16)0X1022)
#define TPM_VEN_ATMEL           ((t_U16)0X1114)
#define TPM_VEN_BROADCOM        ((t_U16)0X14E4)
#define TPM_VEN_IBM             ((t_U16)0X1014)
#define TPM_VEN_INFINEON        ((t_U16)0X15D1)
#define TPM_VEN_INTEL           ((t_U16)0X8086)
#define TPM_VEN_LENOVO          ((t_U16)0X17AA)
#define TPM_VEN_NATIONAL_SEMI   ((t_U16)0X100B)
#define TPM_VEN_QUALCOMM        ((t_U16)0X1011)
#define TPM_VEN_SMSC            ((t_U16)0X1055)
#define TPM_VEN_STMICROELEC     ((t_U16)0X104A)
#define TPM_VEN_SAMSUNG         ((t_U16)0X144D)
#define TPM_VEN_SINOSUN         ((t_U16)0X19FA)
#define TPM_VEN_TEXAS INSTR     ((t_U16)0X104C)
#define TPM_VEN_WINBOND         ((t_U16)0x1050)

/*=== Macro definitions ==============================================================================================*/

/*=== Type definitions for events ====================================================================================*/

/*=== Type definitions ===============================================================================================*/

/*=== Macro-function definitions =====================================================================================*/

/*=== Inline-function definitions ====================================================================================*/

/*=== Function declarations (Interface functions) ====================================================================*/

int tpm_init(void);
int tpm_selftest(void);
t_U32 tpm_getVendorId(void);
int tpm_startup(t_U16 const startupType);
int tpm_nvread(t_U32 const area, t_U32 const offset, t_U8 * const data, t_U32 const dataSize);
int tpm_nvwrite(t_U32 const area, t_U32 const offset, t_U8 const* const data, t_U32 const dataSize);
int tpm_readPcr(t_U32 const index, t_U8* const data);
int tpm_readPcrObscure(t_U32 const index, t_U8* const data);
int tpm_extendPcr(t_U8 const index, t_U8 const * const data, t_U32 const dataSize);
int tpm_resetPcr(t_U8 const index);

#endif
